/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2016                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include<vcg/complex/complex.h>

#include<vcg/complex/algorithms/create/platonic.h>
#include<wrap/io_trimesh/export_off.h>

using namespace vcg;
using namespace std;

class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public UsedTypes<	Use<MyVertex>   ::AsVertexType,
                                        Use<MyEdge>     ::AsEdgeType,
                                        Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public Vertex<MyUsedTypes, vertex::Coord3f, vertex::Normal3f, vertex::Qualityf, vertex::Color4b, vertex::BitFlags  >{};
class MyFace    : public Face< MyUsedTypes,  face::VertexRef, face::FFAdj, face::BitFlags > {};
class MyEdge    : public Edge<MyUsedTypes>{};
class MyMesh    : public tri::TriMesh< vector<MyVertex>, vector<MyFace> , vector<MyEdge>  > {};

typedef  face::Pos<MyFace> MyPos; 

float func(float x, float y)
{
  return cos(sqrt(x*x + y*y)*20.0)/10.0;
}


bool testEdge(MyPos p)
{
  if(p.IsBorder()) return  false;
  
  MyPos tp=p;
  MyVertex *v0=p.V();
  tp.FlipF();tp.FlipE();tp.FlipV();
  MyVertex *v1=tp.V();
  tp.FlipE();tp.FlipV();tp.FlipE();
  MyVertex *v2=tp.V();
  tp.FlipF();tp.FlipE();tp.FlipV();
  MyVertex *v3=tp.V();
  
  float qOld = std::min(Quality(v0->P(),v2->P(),v3->P()),Quality(v0->P(),v1->P(),v2->P()));
  float qNew = std::min(Quality(v0->P(),v1->P(),v3->P()),Quality(v2->P(),v3->P(),v1->P()));
  if(qNew < qOld/1.1) return false;
  Point3f midOld = (v0->P()+v2->P())/2.0f;
  Point3f midNew = (v1->P()+v3->P())/2.0f;
  float fOld = func(midOld.X(),midOld.Y());
  float fNew = func(midNew.X(),midNew.Y());
  float errOld = fabs(midOld.Z()-fOld);
  float errNew = fabs(midNew.Z()-fNew);
  
  return (errNew < errOld);
}

int main( int argc, char **argv )
{
  MyMesh m;
  
  int divNum=40;
  if(argc==2) divNum=atoi(argv[1]);
  float step=1.0/divNum;
  
  std::vector<float> zVec;
  for(int  x=0; x<divNum;++x)
    for(int y=0; y<divNum;++y)
    {
      zVec.push_back(func( (x)/float(divNum), (y)/float(divNum)));
    }
  
  tri::Grid(m,divNum,divNum,1.0,1.0,&(zVec[0]));
  printf("Generated a %i %i grid of %i faces\n",divNum,divNum,m.fn);
  tri::io::ExporterOFF<MyMesh>::Save(m,"grid.off");
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateFlags<MyMesh>::FaceClearV(m);
  // Foreach Edge if swapping improve, swap 
  for(int loop=0;loop<3;++loop)
  {
    int swapCnt=0;
    for(auto fi=m.face.begin();fi!=m.face.end();++fi)
    {
      for(int i=0;i<3;++i)
      {
        MyPos pi(&*fi,i);
        if(!pi.FFlip()->IsV())
          if(testEdge(pi) && face::CheckFlipEdge(*fi,i))
          {
            face::FlipEdge(*fi,i);
            swapCnt++;
          }      
      }
      fi->SetV();
    }
    printf("Performed %i swaps\n",swapCnt);
  }
  tri::io::ExporterOFF<MyMesh>::Save(m,"gridFlip.off");
    
  return 0;
}
